/*
 This is the firmware for the 2006Prog SDA-2006 EEPROM Reader and Programmer.
 This code is copyright (c)2013 by Matthew D'Asaro under the terms of the GPL.
 For usage directions please see included documentation or the on-screen help
 availiable during use.

 Connect the programmer to any PC or Mac with a standard serial port. Use the
 following serial port settings:
 9600 Baud, 8N1, flow control off.
 */

#include <htc.h>
#include <math.h>
#include <stdio.h>
#include "usart.h"
__CONFIG(FOSC_INTOSCIO & CP_OFF & LVP_OFF & CPD_OFF & WRT_OFF & DEBUG_OFF
        & BOREN_OFF & WDTE_OFF & PWRTE_ON & MCLRE_OFF);

//Definitions
#define Size 64 //Size of EEPROM in bytes

unsigned char buffer[Size]={0};
unsigned char verifybuffer[Size]={0};

//Outputs
#define RES RA2
#define DI  RB0
#define CLK RB1
#define REC RA3

//Inputs
#define L   RA0
#define DQ  RA1

//Control Words
//C3 B2 B1 A4 A3 A2 A1 A0
#define readoutcmd 0b01000000
#define programcmd 0b00000000

//Control Characters
#define SOH 1
#define EOT 4


//Functions

//Sleep waits for the specified number of counts. 1 Count ~=36uS
void sleep(long int counts){
    for(long int i=0; i<counts; i++){}
}

//Putch sends one byte out the serial port
void putch(unsigned char data){
	while(!TXIF)	
		continue;
	TXREG = data;
}
//Getch gets one byte but does not echo it back to the sender
unsigned char getch(){
	while(!RCIF){
            //Clear buffer overrun. Characters will be lost but at least PIC
            //won't appear to freeze up.
            if (OERR){
                printf("\r\nSerial buffer overflow. Send data more slowly.\r\n");
                sleep(15000);
                CREN=0;
                CREN=1;
                RCREG=0; //Clear the recieve buffer
                return 0; //Abort buffer load
            }
        }
	return RCREG;
}

//Getche gets one byte and echos it back to the sender unless the byte is backspace.
unsigned char getche(void){	unsigned char data;
        data = getch();
        if(data!='\b'){
            putch(data);
            return data;
        }
        return 0;
}

//Prompt prompts the user and returns the first character of their response.
unsigned char prompt(){
    unsigned char input=0;
    unsigned char command=0;
    printf("2006Prog> ");
    //Get the first byte of entry - this is the command
    do{
        command=getche();
    }while(command==0);

    //If the command is an ESC return it without waiting for a return.
    if(command == 27)
        return command;
    
    //If the command is a return, print a newline and return null
    if(command == '\r'){
        putch('\n');
        return 0;
    }
    //If the command is a page feed return null without printing a newline
    if(command == '\f'){
        return 0;
    }
    
    //Else wait for the user to press return;
    do{
        input=getche();
    }while(input!='\r');
    putch('\n');
    return command;
}

//Help displays the help screen
void help(){
    printf("SDA-2006 EEPROM Reader and Programmer Command Reference:\r\n");
    putch('\n');
    printf("For this screen type 'help' or 'h'\r\n");
    putch('\n');
    printf("To read the EEPROM type 'read' or 'r'\r\n");
    printf("To verify the EEPROM type 'verify' or 'v'\r\n");
    printf("To program the EEPROM type 'program' or 'p'\r\n");
    printf("To erase the EEPROM type 'erase' or 'e'\r\n");
    putch('\n');
    printf("To display the buffer type 'display' or 'd'\r\n");
    printf("To load the buffer manually type 'load' or 'l'\r\n");
    printf("To clear the buffer type 'clear' or 'c'\r\n");
    putch('\n');
    printf("To enter automation mode type 'auto' or 'a' (Reset the system with ESC to exit)\r\n");
    putch('\n');
    printf("Note that only the first letter of a command is read.\r\n");
    printf("For example 'e', 'erase' and 'ehjkdjsgjt' will all erase the EEPROM\r\n");
    putch('\n');
    printf("Press ESC at any time to reset the system (Note that this clears the buffer)\r\n");
}

//Clear clears the buffer
void clear(){
    for(int i=0; i<Size; i++){
        buffer[i]=0;
    }
}
//Clock strobes the clock
void clock(){
    CLK=1;
    sleep(1);
    CLK=0;
}

//Reset resets the EEPROM
void reset(){
    RES=0;
    sleep(1);
    RES=1;
}

//Shiftin sends data to the chip. The LSB on the input char is sent in first
void shiftin(unsigned char input){
    for(int i=0; i<8; i++){
        DI=input>>i;
        clock();
    }
}

//Shiftout recieves 8-bits of data from the chip and returns it
//The first bit shifted out is stored as the LSB
unsigned char shiftout(){
    unsigned char output=0;
    for(int i=0; i<8; i++){
        clock();
        output=output | (DQ<<i);
    }
    return output;
}

//Read reads the contents of the EEPROM into the buffer
void read(){
    clear();
    reset();
    for(unsigned char i=0; i<(Size/2); i++){
        REC=1; //Set REC high
        shiftin(readoutcmd | i); //Shift in the read command and address
        REC=0; //Set REC low - DQ now becomes low impedance
        clock(); //Send clock pulse S to begin readout process
        buffer[2*i+1]=shiftout();
        buffer[2*i]=shiftout();
    }
}

//Program programs the contents of the buffer into the EEPROM
void program(){
    reset();
    for(unsigned char i=0; i<(Size/2); i++){
        REC=1; //Set REC high
        shiftin(buffer[2*i+1]); //Shift in the data
        shiftin(buffer[2*i]);
        shiftin(programcmd | i); //Shift in the program command and address
        REC=0; //Set REC low
        clock(); //Send clock pulse S to begin program process
        while(!L); //Wait for data to be programmed
    }
}

//Erase erases the contents of the EEPROM but leaves the buffer intact
void erase(){
    //Save the current buffer
    for(int i=0;i<Size;i++)
        verifybuffer[i]=buffer[i];
    //Clear the buffer
    clear();
    //Program the chip
    program();
    //Restore the Buffer
    for(int i=0;i<Size;i++)
        buffer[i]=verifybuffer[i];
}

//Load allows the user to enter data into the buffer
int load(int automation){
    unsigned char a=0;
    unsigned char b=0;
    //Get Size bytes of data in hex
    for(int i=0; i<Size; i++){
        //Get the two characters representing one hex byte.
        for(int j=0; j<2; j++){
            a=b;
            //Wait for a valid character
            do{
                b=getch();
                //If we get an ESC reset the system
                if (b==27)
                        return -1;
                //If we get a null, assume buffer overflow and abort load
                if (b==0)
                        return 0;
                //Convert lower case to upper case:
                if (b > 'Z')
                    b=b-32;
            }while(b < '0' || b > 'F' || (b > '9' & b < 'A'));
            //When we have a valid character, print it to the screen
            putch(b);
            //Now decode each character to a number ranging from 0 to 15
            if(b <= '9' )
                b = b -'0';
            else
                b = b - 'A' + 10;
        }
        if(i%4==3 & !automation){
            putch('\r');
            putch('\n');
        }
        else if(!automation)
            putch(' ');
        //We now have the two characters representing a valid hex byte but we
        //still need to combine them and store them in the buffer
        buffer[i]=(a<<4) + b;
    }
    return 0;
}

//Display prints the contents of the buffer to the terminal
void display(int automation){
    for (int i=0; i<Size; i=i+4){
        if(!automation)
            printf(" %02X %02X %02X %02X\r\n", buffer[i], buffer[i+1], buffer[i+2], buffer[i+3]);
        else
            printf("%02X%02X%02X%02X", buffer[i], buffer[i+1], buffer[i+2], buffer[i+3]);
    }
}

//Verify verifies the eprom against the buffer contents
int verify(int automation){
    int fail=0;
    //Save the current buffer
    for(int i=0;i<Size;i++)
        verifybuffer[i]=buffer[i];
    //Read the EEPROM
    read();
    //Compare the chip contents (now in the buffer) with the old buffer
    for(int i=0;i<Size;i++)
        if(verifybuffer[i]!=buffer[i]){
            fail=1;
            break;
        }
    if (fail==1){
        if(!automation){
            printf("Verify Fail! EEPROM contents:\r\n");
            display(0);
        }
        //Restore the buffer
        for(int i=0;i<Size;i++)
            buffer[i]=verifybuffer[i];
        return 1;
    }
    return 0;
}

//Main Program
void main(void){
    //Initilizations
    OSCCON=0b01101100; //Set the internal clock to 4 MHz
    INTCON=0;	// Disable Interrupts
    //Make PORTB outputs with all pins low
    PORTB=0;
    TRISB=0;
    //Make PORTA outputs with all pins low except RST which is high and
                                            //L and DQ which are inputs
    PORTA=0b00000100;
    TRISA=0b00000011;
    init_comms();
    unsigned char command=0;
    unsigned char automation=0;
    unsigned char temp=0;
    unsigned char reset=0;
    printf("\f2006Prog SDA-2006 EEPROM Reader and Programmer\r\n");
    printf("(c)2013 Matthew D'Asaro under the terms of the GPL\r\n");
    printf("Type 'h' for help\r\n");
    while(!automation & !reset){
        command=prompt();
        switch(command){
            case 'h':
                help();
                break;
            case 'r':
                printf("Reading the EEPROM...\r\n");
                read();
                printf("...Done\r\n");
                break;
            case 'l':
                printf("Enter the buffer contents, in hex, one byte at a time.\r\n");
                printf("Invalid characters are ignored\r\n");
                if(load(automation)==-1)
                    reset=1;
                break;
            case 'd':
                printf("Buffer data in hex (big endian - most significant byte first):\r\n");
                display(0);
                break;
            case 'e':
                printf("Erasing the EEPROM...\r\n");
                erase();
                printf("...Done!\r\n");
                break;
            case 'p':
                printf("Programming the EEPROM...\r\n");
                program();
                printf("...Done!\r\n");
                break;
            case 'v':
                printf("Verifiying EEPROM...\r\n");
                if(!verify(automation))
                    printf("Verify OK!\r\n");
                break;
            case 'c':
                clear();
                printf("Buffer Cleared\r\n");
                break;
            case 'a':
                printf("2006Prog Automation Mode (ESC to exit)\r\n");
                automation=1;
                break;
            case 0:
                break;
            case 27:
                reset=1;
                break;
            default :
                printf("Invalid Command\r\n");
        }
    }
    while(automation & !reset){
        command=getch();
        temp=0;
        //wait for enter
        while(temp!='\r'){
            if (command==27 | temp==27){
                reset=1;
                break;
            }
            temp=getch();
        }
        switch(command){
            case 'r':
                read();
                printf("1\r\n");
                break;
            case 'v':
                if(!verify(automation))
                    printf("1\r\n");
                else
                    printf("0\r\n");
                break;
            case 'p':
                program();
                printf("1\r\n");
                break;
            case 'e':
                erase();
                printf("1\r\n");
                break;
            case 'd':
                display(automation);
                printf("\r\n");
                break;
            case 'l':
                printf("1\r\n");
                if(load(automation)==-1)
                    reset=1;
                break;
            case 'c':
                clear();
                printf("1\r\n");
                break;
            case 27: //escape
                reset =1;
                break;
        }
    }
}